import fwk

import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import bisplrep
from scipy.interpolate import bisplev
from scipy.interpolate import interp1d
from scipy.interpolate import RBFInterpolator

class Interpolator:

    def __init__(self, _dartSolver, bnd_wing, bnd_wake, nSections, nDim, nPoints=[500, 25], eps=0.05, k=[5, 5]) -> None:
        self.iSolver = _dartSolver

        self.bndWing = bnd_wing
        self.bndWake = bnd_wake

        self.nDim = nDim

        self.Sections, self.Wing, self.Wake = self.__GetWing(nSections, nDim, nPoints, eps, k)
        self.xxPrev = [[np.zeros(len(self.Sections[iSec][iReg][:,0])) for iReg in range(len(self.Sections[iSec]))] for iSec in range(len(self.Sections))]
        self.DeltaStarPrev = [[np.zeros(len(self.Sections[iSec][iReg][:,0])) for iReg in range(len(self.Sections[iSec]))] for iSec in range(len(self.Sections))]
        self.interpMode = 2 # 0 For Rbf, 1 for 1d interpolation.
        self.tms = fwk.Timers()

    def GetInvBC(self, vSolver, couplIter):
        self.tms['InvToVisc'].start()

        # Extract parameters from the inviscid solver
        # Particular care must be taken with space directions; In the inviscid, the airfoil is in the x,z plane.
        # While setting up the viscous solver, y and z must be inverted. 

        # Wing

        nElm = len(self.bndWing.nodes)
        U_wg, M_wg, Rho_wg = np.zeros((nElm, 3)), np.zeros(nElm), np.zeros(nElm)
        for iNode, n in enumerate(self.bndWing.nodes):
            U_wg[iNode,0] = self.iSolver.U[n.row][0]
            U_wg[iNode,1] = self.iSolver.U[n.row][1]
            U_wg[iNode,2] = self.iSolver.U[n.row][2]
            M_wg[iNode] = self.iSolver.M[n.row]
            Rho_wg[iNode] = self.iSolver.rho[n.row]
        
        # Wake

        nElm = len(self.bndWake.nodes)
        U_wk, M_wk, Rho_wk = np.zeros((nElm, 3)), np.zeros(nElm), np.zeros(nElm)
        for iNode, n in enumerate(self.bndWake.nodes):
            U_wk[iNode,0] = self.iSolver.U[n.row][0]
            U_wk[iNode,1] = self.iSolver.U[n.row][1]
            U_wk[iNode,2] = self.iSolver.U[n.row][2]
            M_wk[iNode] = self.iSolver.M[n.row]
            Rho_wk[iNode] = self.iSolver.rho[n.row]
        
        self.tms['interpolation'].start()
        if self.interpMode == 0:
            Ux = self.__RealToSections(U_wg[:,0], U_wk[:,0])
            Uy = self.__RealToSections(U_wg[:,1], U_wk[:,1])
            Uz = self.__RealToSections(U_wg[:,2], U_wk[:,2])

            Me = self.__RealToSections(M_wg, M_wk)
            Rho = self.__RealToSections(Rho_wg, Rho_wk)
        elif self.interpMode == 1:
            Ux = self.__RbfToSections(U_wg[:,0], U_wk[:,0])
            Uy = self.__RbfToSections(U_wg[:,1], U_wk[:,1])
            Uz = self.__RbfToSections(U_wg[:,2], U_wk[:,2])

            Me = self.__RbfToSections(M_wg, M_wk)
            Rho = self.__RbfToSections(Rho_wg, Rho_wk)
        elif self.interpMode == 2:
            Ux = self.__InterpToSections(U_wg[:,0], U_wk[:,0])
            Uy = self.__InterpToSections(U_wg[:,1], U_wk[:,1])
            Uz = self.__InterpToSections(U_wg[:,2], U_wk[:,2])

            Me = self.__InterpToSections(M_wg, M_wk)
            Rho = self.__InterpToSections(Rho_wg, Rho_wk)
        else:
            raise RuntimeError('Unknown interpolation mode ', self.interpMode)
        self.tms['interpolation'].stop()

        # Find stagnation point
        for iSec in range(len(self.Sections)):

            idxStgUp = np.argmin(np.sqrt(Ux[iSec][0]**2+Uy[iSec][0]**2))
            idxStgLw = np.argmin(np.sqrt(Ux[iSec][1]**2+Uy[iSec][1]**2))

            if np.sqrt(Ux[iSec][0][idxStgUp]**2+Uy[iSec][0][idxStgUp]**2) <= np.sqrt(Ux[iSec][1][idxStgLw]**2+Uy[iSec][1][idxStgLw]**2):
                reg = 0
                idx = idxStgUp
            else:
                reg = 1
                idx = idxStgLw

            xUp, xLw = self.__Remesh(self.Sections[iSec][0][:,0], self.Sections[iSec][1][:,0], idx, reg)
            yUp, yLw = self.__Remesh(self.Sections[iSec][0][:,1], self.Sections[iSec][1][:,1], idx, reg)
            zUp, zLw = self.__Remesh(self.Sections[iSec][0][:,2], self.Sections[iSec][1][:,2], idx, reg)

            vSolver.SetGlobMesh(iSec, 0, xUp, zUp, yUp)
            vSolver.SetGlobMesh(iSec, 1, xLw, zLw, yLw)
            vSolver.SetGlobMesh(iSec, 2, self.Sections[iSec][2][:,0],
                                         self.Sections[iSec][2][:,1],
                                         self.Sections[iSec][2][:,2])
            
            UxUp, UxLw = self.__Remesh(Ux[iSec][0], Ux[iSec][1], idx, reg)
            UyUp, UyLw = self.__Remesh(Uy[iSec][0], Uy[iSec][1], idx, reg)
            UzUp, UzLw = self.__Remesh(Uz[iSec][0], Uz[iSec][1], idx, reg)

            vSolver.SetVelocities(iSec, 0, UxUp, UzUp, UyUp)
            vSolver.SetVelocities(iSec, 1, UxLw, UzLw, UyLw)
            vSolver.SetVelocities(iSec, 2, Ux[iSec][2], Uz[iSec][2], Uy[iSec][2])

            MeUp, MeLw = self.__Remesh(Me[iSec][0], Me[iSec][1], idx, reg)
            vSolver.SetMe(iSec, 0, MeUp)
            vSolver.SetMe(iSec, 1, MeLw)
            vSolver.SetMe(iSec, 2, Me[iSec][2])

            RhoUp, RhoLw = self.__Remesh(Rho[iSec][0], Rho[iSec][1], idx, reg)
            vSolver.SetRhoe(iSec, 0, RhoUp)
            vSolver.SetRhoe(iSec, 1, RhoLw)
            vSolver.SetRhoe(iSec, 2, Rho[iSec][2])

            for iReg in range(3):
                vSolver.SetxxExt(iSec, iReg, self.xxPrev[iSec][iReg])
                vSolver.SetDeltaStarExt(iSec, iReg, self.DeltaStarPrev[iSec][iReg])
        self.tms['InvToVisc'].stop()

    def SetBlowingVel(self, vSolver):
        self.tms['ViscToInv'].start()
        BlowingVel = [[np.zeros(0) for iReg in range(len(self.Sections[iSec]))] for iSec in range(len(self.Sections))]
        xCoord = [[np.zeros(0) for iReg in range(len(self.Sections[iSec]))] for iSec in range(len(self.Sections))]
        for iSec in range(len(self.Sections)):
            for iRegion in range(len(self.Sections[iSec])):
                xCoord[iSec][iRegion] = np.asarray(vSolver.ExtractxCoord(iSec, iRegion))
                BlowingVel[iSec][iRegion] = np.asarray(vSolver.ExtractBlowingVel(iSec, iRegion))
                self.DeltaStarPrev[iSec][iRegion] = vSolver.ExtractDeltaStar(iSec, iRegion)
                self.xxPrev[iSec][iRegion] = vSolver.Extractxx(iSec, iRegion)
    
        if self.nDim == 2:

            # Compute cell centroid x positions (in the viscous mesh)

            xCellsUp = np.zeros(len(xCoord[0][0])-1)
            xCellsUp[0] = (xCoord[0][0][1] - xCoord[0][0][0])/2
            cumDx = xCoord[0][0][1] - xCoord[0][0][0]
            for iCell in range(1, len(xCoord[0][0])-1):
                xCellsUp[iCell] = cumDx + (xCoord[0][0][iCell+1] - xCoord[0][0][iCell])/2
                cumDx += (xCoord[0][0][iCell] - xCoord[0][0][iCell-1]) # Cumulative space between cells

            xCellsLw = np.zeros(len(xCoord[0][1])-1)
            xCellsLw[0] = (xCoord[0][1][1] - xCoord[0][1][0])/2
            cumDx = xCoord[0][1][1] - xCoord[0][0][0]
            for iCell in range(1, len(xCoord[0][1])-1):
                xCellsLw[iCell] = cumDx + (xCoord[0][1][iCell+1] - xCoord[0][1][iCell])/2
                cumDx += (xCoord[0][1][iCell] - xCoord[0][1][iCell-1]) # Cumulative space between cells

            xCellsWk = np.zeros(len(xCoord[0][2])-1)
            xCellsWk[0] = self.Sections[iSec][0][-1, 0] + (xCoord[0][2][1] - xCoord[0][2][0])/2
            cumDx = self.Sections[iSec][0][-1, 0] + xCoord[0][2][1] - xCoord[0][2][0]
            for iCell in range(1, len(xCoord[0][2])-1):
                xCellsWk[iCell] = cumDx + (xCoord[0][2][iCell+1] - xCoord[0][2][iCell])/2
                cumDx += (xCoord[0][2][iCell] - xCoord[0][2][iCell-1]) # Cumulative space between cells

            # Create interpolation functions

            blwSplnUp = interp1d(xCellsUp, BlowingVel[0][0], bounds_error=False, fill_value="extrapolate")
            blwSplnLw = interp1d(xCellsLw, BlowingVel[0][1], bounds_error=False, fill_value="extrapolate")
            blwSplnWk = interp1d(xCellsWk, BlowingVel[0][2], bounds_error=False, fill_value="extrapolate")
            
            # Airfoil (Cell centroids are now computed in the inviscid mesh)
            for iElm, e in enumerate(self.bndWing.tag.elems):

                # Compute cg of element e
                posCg = np.zeros(3)
                for n in e.nodes:
                    posCg[0] += n.pos[0]
                    posCg[1] += n.pos[1]
                    posCg[2] += n.pos[2]
                posCg/=len(e.nodes)
                
                if posCg[1]>=0:
                    self.bndWing.setU(iElm, float(blwSplnUp(posCg[0])))
                else:
                    self.bndWing.setU(iElm, float(blwSplnLw(posCg[0])))
            
            for iElm, e in enumerate(self.bndWake.tag.elems):

                # Compute cg of element e
                posCg = np.zeros(3)
                for n in e.nodes:
                    posCg[0] += n.pos[0]
                    posCg[1] += n.pos[1]
                    posCg[2] += n.pos[2]
                posCg/=len(e.nodes)

                self.bndWake.setU(iElm, float(blwSplnWk(posCg[0])))
            

        elif self.nDim == 3:
            pass
        self.tms['ViscToInv'].stop()


    def RunSolver(self):
        self.iSolver.run()

    def ResetSolver(self):
        self.iSolver.reset()

    def GetCp(self,bnd):
        import dart.utils as floU

        Cp = floU.extract(bnd, self.iSolver.Cp)
        return Cp

    def GetAlpha(self):
        return self.iSolver.pbl.alpha

    def GetMinf(self):
        return self.iSolver.pbl.M_inf
    
    def GetCl(self):
        return self.iSolver.Cl
    
    def GetCm(self):
        return self.iSolver.Cm

    def __RealToSections(self, var, varWk):
        pass

    def __InterpToSections(self, var, varWk):

        varWk = np.reshape(varWk, [len(varWk), 1])
        completeWake = np.hstack((self.Wake, varWk))
        _, idx = np.unique(completeWake[:,:3], axis=0, return_index=True)
        completeWake = completeWake[np.sort(idx)]

        var = np.reshape(var, [len(var), 1])
        completeWing = np.hstack((self.Wing, var))

        wingUp = completeWing[completeWing[:,2]>=0, :]
        wingLw = completeWing[completeWing[:,2]<=0, :]

        mappedVar = [[np.zeros(0) for _ in range(3)] for _ in range(len(self.Sections))]

        for iSec in range(len(mappedVar)):
            for iReg in range(len(mappedVar[iSec])):
                if iReg==0:
                    interpFunc = interp1d(wingUp[:,0], wingUp[:,3])
                if iReg==1:
                    interpFunc = interp1d(wingLw[:,0], wingLw[:,3])
                if iReg==2:
                    interpFunc = interp1d(completeWake[:,0], completeWake[:,3])
                
                mappedVar[iSec][iReg] = interpFunc(self.Sections[iSec][iReg][:,0])

        return mappedVar


    def __RbfToSections(self, var, varWk):

        varWk = np.reshape(varWk, [len(varWk), 1])
        completeWake = np.hstack((self.Wake, varWk))
        _, idx = np.unique(completeWake[:,:3], axis=0, return_index=True)
        completeWake = completeWake[np.sort(idx)]

        var = np.reshape(var, [len(var), 1])
        completeWing = np.hstack((self.Wing, var))

        wingUp = completeWing[completeWing[:,2]>=0]
        wingLw = completeWing[completeWing[:,2]<=0]

        mappedVar = [[np.zeros(0) for _ in range(3)] for _ in range(len(self.Sections))]

        for iSec in range(len(mappedVar)):
            for iReg in range(len(mappedVar[iSec])):
                if iReg==0:
                    mappedVar[iSec][iReg] = RBFInterpolator(wingUp[:,:3], wingUp[:,3], neighbors=10, smoothing=0.1, epsilon=0.6, kernel='gaussian', degree=1)(self.Sections[iSec][iReg])
                    #mappedVar[iSec][iReg] = RBFInterpolator(wingUp[:,:3], wingUp[:,3])(self.Sections[iSec][iReg])
                if iReg==1:
                    mappedVar[iSec][iReg] = RBFInterpolator(wingLw[:,:3], wingLw[:,3], neighbors=10, smoothing=0.1, epsilon=0.6, kernel='gaussian', degree=1)(self.Sections[iSec][iReg])
                    #mappedVar[iSec][iReg] = RBFInterpolator(wingLw[:,:3], wingLw[:,3])(self.Sections[iSec][iReg])
                if iReg==2:
                    mappedVar[iSec][iReg] = RBFInterpolator(completeWake[:,:2], completeWake[:,3], neighbors=10, smoothing=0.1, kernel='linear', degree=0)(self.Sections[iSec][iReg][:,:2])
        return mappedVar


    def __Remesh(self, vectorUp, vectorLw, idx, reg):
        """Rebuilds input vector to match the stagnation point position.
        
        args:
        ----
        - VectorUp/VectorLw upper/lower side vector of the desired variable from x_glob = 0 to x_glob = 1
        - idx : Stagnation point position from leading edge point (x_glob=0)
        - reg : 0/1 indicates on which side the stagnation point is.

        return : vUp/vLw distribution from x_glob=stag to x_glob=1 on upper/lower sides.
        ------
        """

        if reg==0:
            if idx!=0:
                vLw = np.insert(vectorLw, 0, np.flip(vectorUp[1:idx]))
                vUp = np.delete(vectorUp, np.s_[:idx-1])
            else:
                vLw = vectorLw
                vUp = vectorUp
        elif reg==1:
            if idx!=0:
                vUp = np.insert(vectorUp, 0, np.flip(vectorLw[1:idx]))
                vLw = np.delete(vectorLw, np.s_[:idx-1])
            else:
                vUp = vectorUp
                vLw = vectorLw
        else:
            raise RuntimeError('Incorrect region specified during remeshing.')
        
        return vUp, vLw
        

    def __GetWing(self, nSections, nDim, nPoints, eps, k):
        "Extract and sort the node coordinates of the airfoil."

        if nDim == 2 and nSections != 1:
            raise RuntimeError('Cannot create more than 1 section on 2D geometry. Specified:', nSections)
            
        wing = np.zeros((len(self.bndWing.nodes),3))
        wake = np.zeros((len(self.bndWake.nodes),3))

        # Extract coordinates.
        for iNode, n in enumerate(self.bndWing.nodes):
            for iDir in range(len(wing[0,:])):
                wing[iNode,iDir] = n.pos[iDir]

        for iNode, n in enumerate(self.bndWake.nodes):
            for iDir in range(len(wake[0,:])):
                wake[iNode, iDir] = n.pos[iDir]

        if nDim == 2: # 2D case
            wing[:,[1,2]] = wing[:,[2,1]]
            wake[:,[1,2]] = wake[:,[2,1]]
            
        # Separate upper and lower sides

        wingUp = wing[wing[:,2]>=0]
        wingLw = wing[wing[:,2]<=0]

        # Sort upper/lower side and remove duplicates
        wingUp = np.delete(wingUp, wingUp[:,1]>1, 0)
        wingLw = np.delete(wingLw, wingLw[:,1]>1, 0)

        _, idxUp = np.unique(wingUp, axis=0, return_index=True)
        wingUp = wingUp[np.sort(idxUp)]
        _, idxLw = np.unique(wingLw, axis=0, return_index=True)
        wingLw = wingLw[np.sort(idxLw)]

        if nDim == 2:
            arfSplnUp = interp1d(wingUp[:,0], wingUp[:,2], kind='cubic')
            arfSplnLw = interp1d(wingLw[:,0], wingLw[:,2], kind='cubic')

            Sections = [[np.zeros((nPoints[1] if i==2 else nPoints[0], 3)) for i in range(3)]]
            for iReg in range(len(Sections[0])):
                Sections[0][iReg][:,1] = wingUp[0,1]
                if iReg != 2:
                    Sections[0][iReg][:,0] = np.linspace(min(wing[:,0]), max(wing[:,0]), nPoints[0])

                    if iReg == 0:
                        for iPoint, x in enumerate(Sections[0][iReg][:,0]):
                            Sections[0][iReg][iPoint,2] = arfSplnUp(x)
                    elif iReg == 1:
                        for iPoint, x in enumerate(Sections[0][iReg][:,0]):
                            Sections[0][iReg][iPoint,2] = arfSplnLw(x)
                else:
                    Sections[0][iReg][:,0] = np.linspace(min(wake[:,0]), max(wake[:,0]), nPoints[1])

        
        elif nDim == 3:
            sUp = (len(wingUp[:,0])-np.sqrt(2*len(wingUp[:,0])) + len(wingUp[:,0])+np.sqrt(2*len(wingUp[:,0])))/2
            sLw = (len(wingLw[:,0])-np.sqrt(2*len(wingLw[:,0])) + len(wingLw[:,0])+np.sqrt(2*len(wingLw[:,0])))/2

            tckUp = bisplrep(wingUp[:,0], wingUp[:,1], wingUp[:,2], kx=k[0], ky=k[1], s=sUp)
            tckLw = bisplrep(wingLw[:,0], wingLw[:,1], wingLw[:,2], kx=k[0], ky=k[1], s=sLw)

            zDistri = np.linspace(min(wing[:,1]), max(wing[:,1]), nSections)

            zDistri[-1] -= 0.2
            
            Sections = [[np.zeros((nPoints[1] if i==2 else nPoints[0], 3)) for i in range(3)] for _ in range(nSections)]

            for iSec, y in enumerate(zDistri):
                for iReg in range(3):
                    print('Sec', iSec, 'y', y, min(wing[abs(wing[:,1]-y)<eps, 0]), max(wing[abs(wing[:,1]-y)<eps, 0]))
                    print(wing[abs(wing[:,1]-y)<0.02, :])
                    if iReg == 0:
                        Sections[iSec][iReg][:,0] = np.linspace(min(wing[abs(wing[:,1]-y)<eps, 0]), max(wing[abs(wing[:,1]-y)<eps, 0]), nPoints[0])
                    elif iReg == 1:
                        Sections[iSec][iReg][:,0] = np.linspace(min(wing[abs(wing[:,1]-y)<eps, 0]), max(wing[abs(wing[:,1]-y)<eps, 0]), nPoints[0])
                    elif iReg == 2:
                        #print('sec', iSec, 'min', min(wake[:, 0]), 'max', max(wake[:, 0]))
                        Sections[iSec][iReg][:,0] = np.linspace(max(wing[abs(wing[:,1]-y)<eps, 0]), max(wake[:, 0]), nPoints[1])
                    Sections[iSec][iReg][:,1] = y
                    for iPt in range(len(Sections[iSec][iReg][:,0])):
                        if iReg==0:
                            Sections[iSec][iReg][iPt,2] = bisplev(Sections[iSec][iReg][iPt, 0], Sections[iSec][iReg][iPt, 1], tckUp)
                        elif iReg==1:
                            Sections[iSec][iReg][iPt,2] = bisplev(Sections[iSec][iReg][iPt, 0], Sections[iSec][iReg][iPt, 1], tckLw)
                        elif iReg == 2:
                            Sections[iSec][iReg][iPt,2] = 0
                        else:
                            raise RuntimeError("dart::vInterpolator3 l.83 0<=iReg<=2")

        # Add leading edge point
        #Sections.pop()

        for iSec in range(len(Sections)):
            for iReg in range(2):
                Sections[iSec][iReg][0,2] = 0
        
        """fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(wing[:,0], wing[:,1], wing[:,2], s=0.3)
        ax.scatter(wake[:,0], wake[:,1], wake[:,2], s=0.3)
        for iSec in range(len(Sections)):
            ax.scatter(Sections[iSec][0][:,0], Sections[iSec][0][:,1], Sections[iSec][0][:,2], color='red')
            ax.scatter(Sections[iSec][1][:,0], Sections[iSec][1][:,1], Sections[iSec][1][:,2], color='red')
            ax.scatter(Sections[iSec][2][:,0], Sections[iSec][2][:,1], Sections[iSec][2][:,2], color='red')
        ax.set_zlim([-0.5, 0.5])
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.show()

        plt.plot(Sections[0][0][:,0], Sections[0][0][:,2], 'x-', color='red')
        plt.plot(Sections[0][1][:,0], Sections[0][1][:,2], 'x-', color='red')
        plt.plot(wingUp[:,0], wingUp[:,2], 'o', color='blue')
        plt.plot(wingLw[:,0], wingLw[:,2], 'o', color='blue')
        plt.ylim([-0.5, 0.5])
        plt.show()
        #quit()

        for iSec in range(len(Sections)):
            plt.plot(Sections[iSec][0][:,0], Sections[iSec][0][:,2], 'x-', color='red')
            plt.plot(Sections[iSec][1][:,0], Sections[iSec][1][:,2], 'x-', color='red')
            plt.xlim([-0.5, 1.5])
            plt.ylim([-0.5,0.5])
            plt.show()"""
        return Sections, wing, wake



    