import numpy as np
from matplotlib import pyplot as plt
from fwk.coloring import ccolors

# 3D spline with Bézier curves
from scipy.interpolate import bisplrep
from scipy.interpolate import bisplev

# 2D spline with Bézier curves
from scipy.interpolate import splprep
from scipy.interpolate import splev
from scipy.integrate import cumtrapz

from scipy.interpolate import interp1d

from scipy.interpolate import RBFInterpolator

import fwk



class Interpolator:
    def __init__(self, _dartSolver, bnd_wing, bnd_wake, _cfg):
        self.iSolver = _dartSolver

        self.bndWing = bnd_wing
        self.bndWake = bnd_wake
        self.invStruct, self.viscStruct = self.__GetWing(bnd_wing, bnd_wake, _cfg)
        self.config = _cfg

        self.xxPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        self.DeltaStarPrev = [[np.zeros(len(self.viscStruct[iSec][iReg][:,0])) for iReg in range(2)] for iSec in range(len(self.viscStruct))]
        self.tms = fwk.Timers()

    def GetInvBC(self, vSolver, couplIter):
        self.tms['InvToVisc'].start()

        # Extract parameters from the inviscid solver
        # Particular care must be taken with space directions; In the inviscid, the airfoil is in the x,z plane.
        # While setting up the viscous solver, y and z must be inverted. 

        # Wing

        nElm = len(self.bndWing.nodes)
        U_wg, M_wg, Rho_wg = np.zeros((nElm, 3)), np.zeros(nElm), np.zeros(nElm)
        for iNode, n in enumerate(self.bndWing.nodes):
            U_wg[iNode,0] = self.iSolver.U[n.row][0]
            U_wg[iNode,1] = self.iSolver.U[n.row][1]
            U_wg[iNode,2] = self.iSolver.U[n.row][2]
            M_wg[iNode] = self.iSolver.M[n.row]
            Rho_wg[iNode] = self.iSolver.rho[n.row]
        
        # Wake

        nElm = len(self.bndWake.nodes)
        U_wk, M_wk, Rho_wk = np.zeros((nElm, 3)), np.zeros(nElm), np.zeros(nElm)
        for iNode, n in enumerate(self.bndWake.nodes):
            U_wk[iNode,0] = self.iSolver.U[n.row][0]
            U_wk[iNode,1] = self.iSolver.U[n.row][2] # x and y inverted
            U_wk[iNode,2] = self.iSolver.U[n.row][1]
            M_wk[iNode] = self.iSolver.M[n.row]
            Rho_wk[iNode] = self.iSolver.rho[n.row]

        Ux = self.__RbfToSections(U_wg[:,0], U_wk[:,0])
        Uy = self.__RbfToSections(U_wg[:,1], U_wk[:,1])
        Uz = self.__RbfToSections(U_wg[:,2], U_wk[:,2])

        Me = self.__RbfToSections(M_wg, M_wk)
        Rho = self.__RbfToSections(Rho_wg, Rho_wk)

        # Find stagnation point
        for iSec in range(len(self.viscStruct)):

            #if couplIter == 0:
            self.stgPt = np.argmin(Ux[iSec][0]**2+Uy[iSec][0]**2)

            xUp, xLw = self.__Remesh(self.viscStruct[iSec][0][:,0], self.stgPt)
            yUp, yLw = self.__Remesh(self.viscStruct[iSec][0][:,1], self.stgPt)
            zUp, zLw = self.__Remesh(self.viscStruct[iSec][0][:,2], self.stgPt)

            vSolver.SetGlobMesh(iSec, 0, xUp, yUp, zUp)
            vSolver.SetGlobMesh(iSec, 1, xLw, yLw, zLw)
            vSolver.SetGlobMesh(iSec, 2, self.viscStruct[iSec][1][:,0],
                                         self.viscStruct[iSec][1][:,1],
                                         self.viscStruct[iSec][1][:,2])
            
            UxUp, UxLw = self.__Remesh(Ux[iSec][0], self.stgPt)
            UyUp, UyLw = self.__Remesh(Uy[iSec][0], self.stgPt)
            UzUp, UzLw = self.__Remesh(Uz[iSec][0], self.stgPt)

            vSolver.SetVelocities(iSec, 0, UxUp, UyUp, UzUp)
            vSolver.SetVelocities(iSec, 1, UxLw, UyLw, UzLw)
            vSolver.SetVelocities(iSec, 2, Ux[iSec][1], Uz[iSec][1], Uy[iSec][1])

            MeUp, MeLw = self.__Remesh(Me[iSec][0], self.stgPt)

            vSolver.SetMe(iSec, 0, MeUp)
            vSolver.SetMe(iSec, 1, MeLw)
            vSolver.SetMe(iSec, 2, Me[iSec][1])

            RhoUp, RhoLw = self.__Remesh(Rho[iSec][0], self.stgPt)

            vSolver.SetRhoe(iSec, 0, RhoUp)
            vSolver.SetRhoe(iSec, 1, RhoLw)
            vSolver.SetRhoe(iSec, 2, Rho[iSec][1])

            DeltaStarPrevUp, DeltaStarPrevLw = self.__Remesh(self.DeltaStarPrev[iSec][0], self.stgPt)
            vSolver.SetDeltaStarExt(iSec, 0, DeltaStarPrevUp)
            vSolver.SetDeltaStarExt(iSec, 1, DeltaStarPrevLw)
            vSolver.SetDeltaStarExt(iSec, 2, self.DeltaStarPrev[iSec][1])

            xxPrevUp, xxPrevLw = self.__Remesh(self.xxPrev[iSec][0], self.stgPt)
            vSolver.SetxxExt(iSec, 0, xxPrevUp)
            vSolver.SetxxExt(iSec, 1, xxPrevLw)
            vSolver.SetxxExt(iSec, 2, self.xxPrev[iSec][1])
            
            

    def SetBlowingVel(self, vSolver):
        BlowingVel = [[np.zeros(0) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        for iSec in range(len(self.viscStruct)):
            for iRegion in range(3):
                BlowingVel[iSec][iRegion] = np.asarray(vSolver.ExtractBlowingVel(iSec, iRegion))


        if self.config['nDim'] == 2:
            BlwWing = np.concatenate((BlowingVel[0][0][::-1], BlowingVel[0][1]))
            BlwWake = BlowingVel[0][2]
            self.DeltaStarPrev[iSec][0] = np.concatenate((np.asarray(vSolver.ExtractDeltaStar(iSec, 0))[::-1], np.asarray(vSolver.ExtractDeltaStar(iSec, 1))[1:]), axis=None)
            self.DeltaStarPrev[iSec][1] = np.asarray(vSolver.ExtractDeltaStar(iSec, 2))
            self.xxPrev[iSec][0] = np.concatenate((np.asarray(vSolver.Extractxx(iSec, 0))[::-1], np.asarray(vSolver.Extractxx(iSec, 1))[1:]), axis=None)
            self.xxPrev[iSec][1] = np.asarray(vSolver.Extractxx(iSec, 2))

            # Compute cell centers
            viscCgWing = np.zeros((len(self.viscStruct[0][0][:,0]) - 1, 3))
            for iElm in range(len(self.viscStruct[0][0][:,0]) - 1):
                viscCgWing[iElm, 0] = self.viscStruct[0][0][iElm, 0] + self.viscStruct[0][0][iElm + 1, 0]
                viscCgWing[iElm, 1] = self.viscStruct[0][0][iElm, 1] + self.viscStruct[0][0][iElm + 1, 1]
                viscCgWing[iElm, 2] = self.viscStruct[0][0][iElm, 2] + self.viscStruct[0][0][iElm + 1, 2]
            viscCgWing/=2

            viscCgWake = np.zeros((len(self.viscStruct[0][1][:,0]) - 1, 3))
            for iElm in range(len(self.viscStruct[0][1][:,0]) - 1):
                viscCgWake[iElm, 0] = self.viscStruct[0][1][iElm, 0] + self.viscStruct[0][1][iElm + 1, 0]
                viscCgWake[iElm, 1] = self.viscStruct[0][1][iElm, 1] + self.viscStruct[0][1][iElm + 1, 1]
                viscCgWake[iElm, 2] = self.viscStruct[0][1][iElm, 2] + self.viscStruct[0][1][iElm + 1, 2]
            viscCgWake/=2

            invCgWing = np.zeros((len(self.bndWing.tag.elems), 3))
            for iElm, e in enumerate(self.bndWing.tag.elems):
                for n in e.nodes:
                    invCgWing[iElm, 0] +=n.pos[0]
                    invCgWing[iElm, 1] +=n.pos[1]
                    invCgWing[iElm, 2] +=n.pos[2]
                invCgWing[iElm, :] /= e.nodes.size()
            invCgWake = np.zeros((len(self.bndWake.tag.elems), 3))
            for iElm, e in enumerate(self.bndWake.tag.elems):
                for n in e.nodes:
                    invCgWake[iElm, 0] +=n.pos[0]
                    invCgWake[iElm, 1] +=n.pos[1]
                    invCgWake[iElm, 2] +=n.pos[2]
                invCgWake[iElm, :] /= e.nodes.size()
            
            """plt.plot(self.invStruct[0][:,0], self.invStruct[0][:,1], 'x')
            plt.plot(invCgWing[:, 0], invCgWing[:,1], 'o')
            plt.show()"""

        self.tms['Interpolation'].start()
        mappedBlwWing = RBFInterpolator(viscCgWing, BlwWing, neighbors=5, kernel='linear', smoothing=1e-6)(invCgWing)
        mappedBlwWake = RBFInterpolator(viscCgWake, BlwWake, neighbors=5, kernel='linear', smoothing=1e-6)(invCgWake)
        self.tms['Interpolation'].stop()
        for iElm in range(len(self.bndWing.tag.elems)):
            self.bndWing.setU(iElm, mappedBlwWing[iElm])
        for iElm in range(len(self.bndWake.tag.elems)):
            self.bndWake.setU(iElm, mappedBlwWake[iElm])


    def RunSolver(self):
        self.iSolver.run()

    def ResetSolver(self):
        self.iSolver.reset()

    def GetCp(self,bnd):
        import dart.utils as floU

        Cp = floU.extract(bnd, self.iSolver.Cp)
        return Cp

    def GetAlpha(self):
        return self.iSolver.pbl.alpha

    def GetMinf(self):
        return self.iSolver.pbl.M_inf
    
    def GetCl(self):
        return self.iSolver.Cl
    
    def GetCm(self):
        return self.iSolver.Cm

    def __Remesh(self, vec, idx_stag):
        xUp = vec[:idx_stag+1]
        xUp = xUp[::-1]

        xLw = vec[idx_stag:]
        return xUp, xLw

    def __RbfToSections(self, var, varWk):
        varWk = np.reshape(varWk, [len(varWk), 1])
        completeWake = np.hstack((self.invStruct[1], varWk))
        _, idx = np.unique(completeWake[:,:3], axis=0, return_index=True)
        completeWake = completeWake[np.sort(idx)]

        var = np.reshape(var, [len(var), 1])
        completeWing = np.hstack((self.invStruct[0], var))

        mappedVar = [[np.zeros(0) for _ in range(3)] for _ in range(len(self.viscStruct))]

        for iSec in range(len(mappedVar)):
            for iReg in range(len(mappedVar[iSec])):
                if iReg==0: # Wing
                    mappedVar[iSec][iReg] = RBFInterpolator(completeWing[:,:3], completeWing[:,3], neighbors=5, kernel='linear', smoothing=1e-2)(self.viscStruct[iSec][0])
                if iReg==1: # Wake
                    mappedVar[iSec][iReg] = RBFInterpolator(completeWake[:,:3], completeWake[:,3], neighbors=5, kernel='linear', smoothing=1e-2)(self.viscStruct[iSec][1])
        
        """plt.plot(completeWing[:,0], completeWing[:,3], 'x')
        plt.plot(completeWake[:,0], completeWake[:,3], 'x')
        plt.plot(self.viscStruct[0][0][:,0], mappedVar[0][0])
        plt.plot(self.viscStruct[0][1][:,0], mappedVar[0][1])
        plt.show()"""
        return mappedVar


    def __GetWing(self, bnd_wing, bnd_wake, config):

        if config['nDim'] == 2 and config['nSections'] != 1:
            raise RuntimeError('Cannot create more than 1 section on 2D geometry. Specified:', config['nSections'])

        viscStruct = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0], 3)) for i in range(2)] for _ in range(config['nSections'])]

        invStruct = [np.zeros((len(bnd_wake.nodes) if iReg==1 else len(bnd_wing.nodes), 3)) for iReg in range(2)]

        for iNode, n in enumerate(bnd_wing.nodes):
            invStruct[0][iNode, 0] = n.pos[0]
            invStruct[0][iNode, 1] = n.pos[1]
            invStruct[0][iNode, 2] = n.pos[2]
        for iNode, n in enumerate(bnd_wake.nodes):
            invStruct[1][iNode, 0] = n.pos[0]
            invStruct[1][iNode, 1] = n.pos[2]
            invStruct[1][iNode, 2] = n.pos[1]

        if config['nDim'] == 2:
            spanDistri = np.zeros(1)
            spanDistri[0] = invStruct[0][0,2] # Span direction is the z direction 

        elif config['nDim'] == 3:
            invStruct[0][:,[1,2]] = invStruct[0][:,[2,1]] # Inverse y and z so that spanwise direction is in z
            invStruct[1][:,[1,2]] = invStruct[1][:,[2,1]] # Inverse y and z so that spanwise direction is in z
            spanDistri = np.linspace(min(invStruct[0][:,2]), max(invStruct[0][:,2]), config['nSections'])
            spanDistri[0] +=0.1
            spanDistri[-1] -=0.1

        for iSec, z in enumerate(spanDistri):

            # Create x distribution at the corresponding span position
            minX = min(invStruct[0][abs(invStruct[0][:,2]-z)<=config['eps'], 0])
            chordLength = max(invStruct[0][abs(invStruct[0][:,2]-z)<=config['eps'], 0]) - minX
            xDistri = minX + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength
            nElmSide = len(xDistri)
            viscStruct[iSec][0][:,0] = np.concatenate((xDistri, xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            wakeLength = np.max(invStruct[1][:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))

            if config['nDim'] == 2:
                upper, lower = self.__SeparateArf(invStruct[0][:,0], invStruct[0][:,1])
                viscStruct[iSec][0][:nElmSide,1] = interp1d(upper[:,0], upper[:,1])(viscStruct[iSec][0][:nElmSide,0])
                viscStruct[iSec][0][nElmSide:,1] = interp1d(lower[:,0], lower[:,1])(viscStruct[iSec][0][nElmSide:,0])
                viscStruct[iSec][0] = np.delete(viscStruct[iSec][0], nElmSide, axis=0)

                viscStruct[iSec][1][:,1] = interp1d(invStruct[1][:,0], invStruct[1][:,1])(viscStruct[iSec][1][:,0])

                """plt.plot(viscStruct[iSec][0][:,0], viscStruct[iSec][0][:,1], '-')
                plt.plot(viscStruct[iSec][0][:nElmSide,0], viscStruct[iSec][0][:nElmSide,1], 'x')
                plt.plot(viscStruct[iSec][0][nElmSide:,0], viscStruct[iSec][0][nElmSide:,1], 'o')
                plt.plot(viscStruct[iSec][1][:,0], np.zeros(len(viscStruct[iSec][1][:,0])), 'x-')
                plt.xlim([-0.5,2])
                plt.ylim([-0.5, 0.5])
                plt.show()"""

            elif config['nDim'] == 3:
                nPointsMin = (config['k'][0]+1) * (config['k'][1]+1) # See Scipy bisplrep doc

                wingSpan = max(invStruct[0][:,2]) - min(invStruct[0][:,2])
                portion = invStruct[0][abs(invStruct[0][:,2] - z) < 0.01 * wingSpan,:]

                portionUpper, portionLower = self.__SeparateArf(portion[:,0], portion[:,1], portion[:,2])
                if len(portionUpper)<=nPointsMin or len(portionLower)<=nPointsMin:
                    print(ccolors.ANSI_YELLOW,'dart::Interpolator Warning: Found {:.0f} points within 1% of spanwise position z={:.2f}. Trying 10%. '.format(
                        len(portion), z),ccolors.ANSI_RESET)
                    portion = invStruct[0][abs(invStruct[0][:,2] - z) < 0.1 * wingSpan,:]
                    if len(portion)<=nPointsMin:
                        print(ccolors.ANSI_RED, 'Failed. Continuing\n', ccolors.ANSI_RESET)
                    else:
                        print(ccolors.ANSI_GREEN, 'Sucess.\n', ccolors.ANSI_RESET)

                # Upper
                sWing = (len(portionUpper[:,0])-np.sqrt(2*len(portionUpper[:,0])) + len(portionUpper[:,0])+np.sqrt(2*len(portionUpper[:,0])))/2
                tckWingUp = bisplrep(portionUpper[:,0], portionUpper[:,2], portionUpper[:,1], kx=config['k'][0], ky=config['k'][1], s=sWing)
                for iPoint in range(nElmSide):
                    viscStruct[iSec][0][iPoint,1] = bisplev(viscStruct[iSec][0][iPoint,0], viscStruct[iSec][0][iPoint,2], tckWingUp)
                
                # Lower
                sWing = (len(portionLower[:,0])-np.sqrt(2*len(portionLower[:,0])) + len(portionLower[:,0])+np.sqrt(2*len(portionLower[:,0])))/2
                tckWingUp = bisplrep(portionLower[:,0], portionLower[:,2], portionLower[:,1], kx=config['k'][0], ky=config['k'][1], s=sWing)
                for iPoint in range(nElmSide, len(viscStruct[iSec][0][:,0])):
                    viscStruct[iSec][0][iPoint,1] = bisplev(viscStruct[iSec][0][iPoint,0], viscStruct[iSec][0][iPoint,2], tckWingUp)

                # Wake
                sWake = (len(invStruct[1][:,0])-np.sqrt(2*len(invStruct[1][:,0])) + len(invStruct[1][:,0])+np.sqrt(2*len(invStruct[1][:,0])))/2
                tckWingUp = bisplrep(invStruct[1][:,0], invStruct[1][:,2], invStruct[1][:,1], kx=config['k'][0], ky=config['k'][1], s=sWake)
                for iPoint in range(len(viscStruct[iSec][1][:,0])):
                    viscStruct[iSec][1][iPoint,1] = bisplev(viscStruct[iSec][1][iPoint,0], viscStruct[iSec][1][iPoint,2], tckWingUp)


        if config['nDim'] ==3:
            fig = plt.figure()
            ax = fig.add_subplot(projection='3d')
            ax.scatter(invStruct[0][:,0], invStruct[0][:,1], invStruct[0][:,2], s=0.2)
            for iSec in range(len(viscStruct)):
                ax.scatter(viscStruct[iSec][0][:,0], viscStruct[iSec][0][:,1], viscStruct[iSec][0][:,2], s=0.9, color='red')
            ax.set_ylim([-0.5, 0.5])
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            plt.show()
        return invStruct, viscStruct

    def __SeparateArf(self, xInp, yInp, zInp=None):
        if zInp is None:
            zInp = np.zeros(len(xInp))
        if len(xInp) != len(yInp) or len(xInp) != len(zInp):
            raise RuntimeError ('dart::Interpolator::__SeparateArf Missmatch between intput vector sizes.')

        Q=np.zeros((len(xInp), 3))
        Q[:,0] = xInp.copy()
        Q[:,1] = yInp.copy()
        Q[:,2] = zInp.copy()

        Q = Q[Q[:, 0].argsort()]

        le = Q[0,:]
        te = Q[-1,:]
        ymean = (le[1] + te[1]) / 2

        upper, lower, save = le, le, le

        flag1 = 0
        iPoint = 0
        for i in range(len(Q[:,0])):
            if Q[iPoint,1] == 0.0:
                if flag1 == 1:
                    Q=np.delete(Q, iPoint, axis = 0)
                    iPoint-=1
                else:
                    flag1 = 1
            iPoint +=1

        for iPoint in range(1, len(Q[:,0])):
            if Q[iPoint,1] > ymean:
                upper = np.vstack((upper, Q[iPoint,:]))
            elif Q[iPoint, 1] < ymean:
                lower = np.vstack((lower, Q[iPoint,:]))
            else:
                save = np.vstack((save, Q[iPoint, :]))
        
        _, idx = np.unique(save, return_index=True, axis=0)
        save=save[np.sort(idx)]

        #save = np.delete(save, 0, axis=0)

        
        try:
            if len(save[:,0])>0:
                upper = np.vstack((upper, save))
                lower = np.vstack((lower, save))
        except:
            pass

        upper = np.vstack((upper, te))
        lower = np.vstack((lower, te))

        return upper, lower