#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute adjoint solution of lifting (linear or nonlinear) potential flow around a NACA 0012
# Adrien Crovato
#
# Test the adjoint solver
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import numpy as np
import dart.utils as floU
import dart.default as floD
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 2*np.pi/180
    M_inf = 0.7
    c_ref = 1
    dim = 2

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.0075, 'msLe' : 0.0075}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    # create mesh deformation handler
    morpher = floD.morpher(msh, dim, ['airfoil'])
    tms['msh'].stop()

    # set the problem and add fluid, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, c_ref, c_ref, 0.25, 0., 0., 'airfoil', te = 'te')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # solve adjoint problem
    print(ccolors.ANSI_BLUE + 'PyGradient...' + ccolors.ANSI_RESET)
    tms['adjoint'].start()
    adjoint = floD.adjoint(solver, morpher)
    adjoint.run()
    adjoint.save(gmshWriter)
    tms['adjoint'].stop()

    # compute norm of gradient wrt to mesh
    dClX = 0
    dCdX = 0
    for n in msh.nodes:
        dClX += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1]]).dot(np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1]]))
        dCdX += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1]]).dot(np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1]]))
    dClX = np.sqrt(dClX)
    dCdX = np.sqrt(dCdX)

    # recover gradient wrt to AoA from gradient wrt mesh
    drot = np.array([[-np.sin(alpha), np.cos(alpha)], [-np.cos(alpha), -np.sin(alpha)]])
    dClAoA = 0
    dCdAoA = 0
    for n in bnd.nodes:
        dx = drot.dot(np.array([n.pos[0] - 0.25, n.pos[1]]))
        dClAoA += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1]]).dot(dx)
        dCdAoA += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1]]).dot(dx)

    # extract Cp and drag sensitivities
    Cp = floU.extract(bnd.groups[0].tag.elems, solver.Cp)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    dCd = floU.extract(bnd.groups[0].tag.elems, adjoint.dCdMsh)
    tboxU.write(dCd, 'dCd_airfoil.dat', '%1.5e', ', ', 'x, y, z, dCd_dx, dCd_dy, dCd_dz', '')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('       M    alpha       Cl       Cd   dCl_da   dCd_da')
    print('{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f} {5:8.4f}'.format(M_inf, alpha*180/np.pi, solver.Cl, solver.Cd, adjoint.dClAoa, adjoint.dCdAoa))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if M_inf == 0. and alpha == 0*np.pi/180:
        tests.add(CTest('dCl_dAoA', adjoint.dClAoa, 6.9, 1e-2))
        tests.add(CTest('dCd_dAoA', adjoint.dCdAoa, 0.0, 1e-3))
        tests.add(CTest('dCl_dMsh', dClX, 55.439, 1e-3))
        tests.add(CTest('dCd_dMsh', dCdX, 0.482, 1e-3))
        tests.add(CTest('dCl_dAoA (msh)', dClAoA, 6.9, 1e-2))
        tests.add(CTest('dCd_dAoA (msh)', dCdAoA, 0.0, 1e-3))
    elif M_inf == 0.7 and alpha == 2*np.pi/180:
        tests.add(CTest('dCl_dAoA', adjoint.dClAoa, 11.8, 1e-2)) # FD 11.835 (step = 1e-5)
        tests.add(CTest('dCd_dAoA', adjoint.dCdAoa, 0.127, 1e-2)) # 0.12692 (step = 1e-5)
        tests.add(CTest('dCl_dMsh', dClX, 82.67, 1e-3))
        tests.add(CTest('dCd_dMsh', dCdX, 0.908, 1e-3))
        tests.add(CTest('dCl_dAoA (msh)', dClAoA, 11.8, 1e-2))
        tests.add(CTest('dCd_dAoA (msh)', dCdAoA, 0.127, 1e-2))
    else:
        raise Exception('Test not defined for this flow')
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
